from collections import namedtuple
import pandas as pd
import nest

from .nest_utils import nest_version

PopulationID = namedtuple('PopulationID', 'node_id population')


def ids2list_nest2(nest_ids):
    return nest_ids


def ids2list_nest3(nest_ids):
    if isinstance(nest_ids, nest.NodeCollection):
        return nest_ids.tolist()
    else:
        return nest_ids


ids2list = ids2list_nest3 if nest_version[0] >= 3 else ids2list_nest2


class GidPool(object):
    def __init__(self):
        # self._popid2gid = {}  # (pop_name, node_id) --> nest_id
        self._gid2pop_id = {}  # nest_id --> (pop_name, node_id)
        self._nestid_lu = {}

    @property
    def gids(self):
        return list(self._gid2pop_id.keys())

    @property
    def populations(self):
        return list(self._nestid_lu.keys())

    def add(self, name, node_id, gid):
        raise NotImplementedError()

    def get_gid(self, name, node_id):
        return self.get_nestids(name=name, node_ids=[node_id])[0]

    def get_pool_id(self, gid):
        return self._gid2pop_id[gid]

    def create_pool(self, name):
        pass

    def add_nestids(self, name, node_ids, nest_ids):
        # in NEST 3.0+ nest.Create() returns a NodeCollection instead of a list of ids, need to convert
        nest_ids = ids2list(nest_ids)

        if name not in self._nestid_lu:
            lu_table = pd.DataFrame({'nest_ids': nest_ids, 'node_ids': node_ids})
            lu_table = lu_table.set_index('node_ids')
        else:
            new_df = pd.DataFrame({'nest_ids': nest_ids, 'node_ids': node_ids})
            new_df = new_df.set_index('node_ids')
            lu_table = self._nestid_lu[name]
            lu_table = pd.concat((lu_table, new_df))
            # lu_table = lu_table.reindex(lu_table.index.values)

        self._nestid_lu[name] = lu_table

        for node_id, nest_id in zip(node_ids, nest_ids):
            self._gid2pop_id[nest_id] = PopulationID(population=name, node_id=node_id)

    def add_gids(self, name, node_ids, gids):
        self.add_nestids(name=name, node_ids=node_ids, nest_ids=gids)

    def get_nestids(self, name, node_ids):
        nestids_table = self._nestid_lu[name]
        return nestids_table.loc[node_ids]['nest_ids'].values

    def get_gids(self, name, node_ids):
        return self.get_nestids(name=name, node_ids=node_ids)

    def __len__(self):
        return len(self.gids)




class SimInput(object):
    """A helper class for parsing the "inputs" section of a SONATA config file. Separate the actual parameters needed
    to instantiate a module from the metadata parameters (eg, name, module, input_type)

    Use the build() method to parse the json/dictionary of a section of input, then **params** to get a dictionary
    of the values used to instantiate::

        input = SimInput.Build(
                  'spikes_inputs',
                  {'module': 'spikes', 'input_type': 'hdf5', input_file: 'my_spikes.h5', 'time_scale': 'ms'}
        )

        if input.module == 'spike':
            SpikesInput(**params)
            ...

    Attributes:

        * **name** - str, name of module
        * **input_type** - str
        * **module** - str,
        * **params** - dictionary, all parameters (not including name, input_type, module)

    Custom Modules:

        Sometimes certain input types may require extra steps in processing, like auto-conversion of filling in missing
        parameters. In this case use the register module method::

            class MyVClampInput(SimInput):
                def avail_module():
                    return ['vclamp', 'voltage_clamp']

                def build():
                    ....

            SimInput.registerModule(MyVClampInput)

        Then when SimInput.build() is called and the 'module_name'=='vclamp' (or 'voltage_clamp') it will pass the parsing
        to the MyVClampInput class.
    """
    registry = {}  # For factory function

    def __init__(self, input_name, input_type, module, params):
        self.name = input_name
        self.input_type = input_type
        self.module = module
        self.params = params.copy()

        # Remove the 'module' and 'input_type' from the params since user should access it through the variable
        for param_key in ['module', 'input_type']:
            if param_key in self.params:
                del self.params[param_key]

        # Special variable, not a part of standard but still want for ease of testing
        if 'enabled' in params:
            self.enabled = params['enabled']
            del self.params['enabled']
        else:
            self.enabled = True

        # Fill in missing values with default (as specified by the subclass)
        for var_name, default_val in self._get_defaults():
            if var_name not in self.params:
                self.params[var_name] = default_val

        # Check there are no missing parameters

    @property
    def node_set(self):
        return self.params.get('node_set', None)

    def _get_defaults(self):
        return []

    @classmethod
    def build(cls, input_name, params):
        """Creates a SimInput object with parsed out parameters

        :param input_name: name of specific input
        :param params: dictionary of input parameters
        :return: SimInput object, or subclass that matches the specified 'module' value
        """
        params = params.copy()
        if 'module' not in params:
            raise Exception('inputs setting {} does not specify the "module".'.format(input_name))

        if 'input_type' not in params:
            raise Exception('inputs setting {} does not specify the "input_type".'.format(input_name))

        module_name = params['module']
        input_type = params['input_type']
        module_cls = SimInput.registry.get(module_name, SimInput)

        return module_cls(input_name, input_type, module_name, params)

    @classmethod
    def register_module(cls, subclass):
        """

        :param subclass:
        """
        # For factory, register subclass based on the module name(s)
        assert(issubclass(subclass, cls))
        mod_registry = cls.registry
        mod_list = subclass.avail_modules()
        modules = mod_list if isinstance(mod_list, list) else [mod_list]
        for mod_name in modules:
            if mod_name in mod_registry:
                raise Exception('Multiple modules named {}'.format(mod_name))
            mod_registry[mod_name] = subclass

        return subclass


def from_config(cfg):
    """Takes in a bmtk.utils.Config instance and will automatically parse each "input" in the config sections,
    returning a list of SimInput objects. If an input has "enabled" = False then it will automatically be excluded.

    :param cfg: A SONATAConfig object
    :return: A list of SimInput modules corresponding to the parsed inputs of a config.
    """
    inputs_list = []
    for input_name, input_params in cfg.inputs.items():
        input_setting = SimInput.build(input_name, input_params)
        if input_setting.enabled:
            inputs_list.append(input_setting)

    return inputs_list






import numpy as np
import scipy.signal as spsig

from .utilities import convert_tmin_tmax_framerate_to_trange


class KernelCursor(object):
    """A class that takes care of the convolution of the (non-separable) spatial-temporal linear filter with the move.
    """
    def __init__(self, kernel, movie):
        self.movie = movie
        self.kernel = kernel
        self.cache = {}

        # This ensures that the kernel frame rate matches the movie frame rate:
        np.testing.assert_almost_equal(np.diff(self.kernel.t_range),
                                       np.ones_like(self.kernel.t_range[1:])*(1./movie.frame_rate))
        
    @property
    def row_range(self):
        return self.movie.row_range
        
    @property
    def col_range(self):
        return self.movie.col_range
    
    @property
    def t_range(self):
        return self.movie.t_range
    
    @property
    def frame_rate(self):
        return self.movie.frame_rate
    
    def evaluate(self, t_min=None, t_max=None, downsample=1):
        if t_max is None:
            t_max = self.t_range[-1]
            
        if t_min is None:
            t_min = self.t_range[0]
        
        t_range = convert_tmin_tmax_framerate_to_trange(t_min, t_max, self.movie.frame_rate)[::int(downsample)]
        y_vals = np.array([self(t) for t in t_range])

        return t_range, y_vals  
    
    def __call__(self, t):
        # TODO: Using call is not a good idea here, change to evaluate
        if t < self.t_range[0] or t > self.t_range[-1]:
            curr_rate = 0
        else:
            ti = t*self.frame_rate
            til, tir = int(np.floor(ti)), int(np.ceil(ti))
            
            tl, tr = float(til)/self.frame_rate, float(tir)/self.frame_rate
            if np.abs(tl - t) < 1e-12:
                curr_rate = self.apply_dot_product(til)

            elif np.abs(tr - t) < 1e-12:
                curr_rate = self.apply_dot_product(tir)

            else:
                wa, wb = (1-(t-tl)/(tr-tl)), (1-(tr-t)/(tr-tl))
                cl = self.apply_dot_product(til)
                cr = self.apply_dot_product(tir)             
                curr_rate = cl*wa+cr*wb

        if np.isnan(curr_rate):
            assert RuntimeError
        
        return curr_rate

    def apply_dot_product(self, ti_offset):
        try:
            # TODO: This needs to be altered asap
            return self.cache[ti_offset]
        
        except KeyError:
            t_inds = self.kernel.t_inds + ti_offset + 1  # Offset by one nhc 14 Apr '17
            min_ind, max_ind = 0, self.movie.data.shape[0]
            allowed_inds = np.where(np.logical_and(min_ind <= t_inds, t_inds < max_ind))
            t_inds = t_inds[allowed_inds]
            row_inds = self.kernel.row_inds[allowed_inds]
            col_inds = self.kernel.col_inds[allowed_inds]
            kernel_vector = self.kernel.kernel[allowed_inds] 
            result = np.dot(self.movie[t_inds, row_inds, col_inds], kernel_vector)
            self.cache[ti_offset] = result
            return result


class FilterCursor(KernelCursor):
    def __init__(self, spatiotemporal_filter, movie, threshold=0):
        # TODO: not sure why this needs to have it's own class and shouldn't be merged into parent?
        self.spatiotemporal_filter = spatiotemporal_filter
        kernel = self.spatiotemporal_filter.get_spatiotemporal_kernel(movie.row_range, movie.col_range,
                                                                      t_range=movie.t_range, threshold=threshold,
                                                                      reverse=True)

        super(FilterCursor, self).__init__(kernel, movie)


class LNUnitCursor(KernelCursor):
    """A class that takes care of applying a linear-nonlinear filter to a movie. Parent class is used to apply the
    spatial-termporal filter convolution to the movie, when then the lnunit non-linear transfer function is applied.

    """
    def __init__(self, lnunit, movie, threshold=0):
        self.lnunit = lnunit
        kernel = lnunit.get_spatiotemporal_kernel(movie.row_range, movie.col_range, movie.t_range, reverse=True,
                                                  threshold=threshold)
        kernel.apply_threshold(threshold)
             
        super(LNUnitCursor, self).__init__(kernel, movie)
    
    def __call__(self, t):
        # TODO: Don't use call operator, change to evaluate
        return self.lnunit.transfer_function(super(LNUnitCursor, self).__call__(t))


class MultiLNUnitCursor(object):
    def __init__(self, multi_lnunit, movie, threshold=0):
        self.multi_lnunit = multi_lnunit
        self.lnunit_cursor_list = [LNUnitCursor(lnunit, movie, threshold=threshold)
                                   for lnunit in multi_lnunit.lnunit_list]
        self.movie = movie
        
    def evaluate(self, **kwargs):
        multi_e = [unit_cursor.evaluate(**kwargs) for unit_cursor in self.lnunit_cursor_list]
        t_list, y_list = zip(*multi_e)
        return t_list[0], self.multi_lnunit.transfer_function(*y_list)


class MultiLNUnitMultiMovieCursor(MultiLNUnitCursor):
    def __init__(self, multi_lnunit, movie_list, threshold=0.):
        assert len(multi_lnunit.lnunit_list) == len(movie_list)
        
        self.multi_lnunit = multi_lnunit
        self.lnunit_movie_list = movie_list
        self.lnunit_cursor_list = [lnunit.get_cursor(movie, threshold=threshold) for
                                   lnunit, movie in zip(multi_lnunit.lnunit_list, movie_list)]


class SeparableKernelCursor(object):
    """A class for applying a spatial-temporal convolution to a movie. Unlike the KernelCursor the spatial-temporal
    filter is broken up into its components, spatial filter is applied then followed by a temporal filter convolution.

    """
    def __init__(self, spatial_kernel, temporal_kernel, movie):
        """Assumes temporal kernel is not reversed"""

        self.movie = movie
        self.spatial_kernel = spatial_kernel
        self.temporal_kernel = temporal_kernel

    def evaluate(self, threshold=0):
        full_spatial_kernel = np.array([self.spatial_kernel.full()])
        full_temporal_kernel = self.temporal_kernel.full()

        # Convolve every frame in the movie with the spatial filter. First find the range of rows (range min and max)
        #  and columns in the filter that are above threshold, that way only portion of movie/filter are multiplied
        # together
        nonzero_inds = np.where(np.abs(full_spatial_kernel[0, :, :]) >= threshold)
        rm, rM = nonzero_inds[0].min(), nonzero_inds[0].max()
        cm, cM = nonzero_inds[1].min(), nonzero_inds[1].max()
        convolution_answer_sep_spatial = (self.movie.data[:, rm:rM+1, cm:cM+1] *
                                          full_spatial_kernel[:, rm:rM+1, cm:cM+1]).sum(axis=1).sum(axis=1)

        # Convolve results of spatial convolution with the temporal filter
        sig_tmp = np.zeros(len(full_temporal_kernel) + len(convolution_answer_sep_spatial) - 1)
        sig_tmp[len(full_temporal_kernel)-1:] = convolution_answer_sep_spatial
        convolution_answer_sep = spsig.convolve(sig_tmp, full_temporal_kernel[::-1], mode='valid')
        t = np.arange(len(convolution_answer_sep))/self.movie.frame_rate

        return t, convolution_answer_sep


class SeparableSpatioTemporalFilterCursor(SeparableKernelCursor):
    def __init__(self, spatiotemporal_filter, movie):
        self.spatial_filter = spatiotemporal_filter.spatial_filter
        self.temporal_filter = spatiotemporal_filter.temporal_filter

        spatial_kernel = self.spatial_filter.get_kernel(movie.row_range, movie.col_range, threshold=-1)
        temporal_kernel = self.temporal_filter.get_kernel(t_range=movie.t_range, threshold=0, reverse=True)
        spatial_kernel.kernel *= spatiotemporal_filter.amplitude

        super(SeparableSpatioTemporalFilterCursor, self).__init__(spatial_kernel, temporal_kernel, movie)


class SeparableLNUnitCursor(SeparableSpatioTemporalFilterCursor):
    def __init__(self, lnunit, movie):
        self.lnunit = lnunit
        super(SeparableLNUnitCursor, self).__init__(self.lnunit.linear_filter, movie)

    def evaluate(self, downsample=1):
        assert(downsample == 1)
        t, y = super(SeparableLNUnitCursor, self).evaluate()
        return t, [self.lnunit.transfer_function(yi) for yi in y]


class SeparableMultiLNUnitCursor(object):
    def __init__(self, multilnunit, movie):
        self.multilnunit = multilnunit
        self.lnunit_cursor_list = []
        for lnunit in self.multilnunit.lnunit_list:
            self.lnunit_cursor_list.append(SeparableLNUnitCursor(lnunit, movie))

    def evaluate(self, *args, **kwargs):
        assert(kwargs.get('downsample', 1) == 1)
        y_list = []
        for cursor in self.lnunit_cursor_list:
            t, y = cursor.evaluate(*args, **kwargs)
            y_list.append(y)

        return t, self.multilnunit.transfer_function(*y_list)



# Copyright 2017. Allen Institute. All rights reserved
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
# disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
# disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
import numpy as np
from neuron import h


pc = h.ParallelContext()  # object to access MPI methods


class Morphology(object):
    """Methods for processing morphological data"""
    def __init__(self, hobj):
        """reuse hoc object from one of the cells which share the same morphology/model"""
        self.hobj = hobj
        self.sec_type_swc = {'soma': 1, 'somatic': 1,  # convert section name and section list names
                             'axon': 2, 'axonal': 2,  # into a consistent swc notation
                             'dend': 3, 'basal': 3,
                             'apic': 4, 'apical': 4}
        self.nseg = self.get_nseg()
        self._segments = {}

    def get_nseg(self):
        nseg = 0
        for sec in self.hobj.all:
            nseg += sec.nseg  # get the total # of segments in the cell
        return nseg

    def get_soma_pos(self):
        n3dsoma = 0
        r3dsoma = np.zeros(3)
        for sec in self.hobj.soma:
            n3d = int(h.n3d(sec=sec))  # get number of n3d points in each section
            r3d = np.zeros((3, n3d))  # to hold locations of 3D morphology for the current section
            n3dsoma += n3d

            for i in range(n3d):
                r3dsoma[0] += h.x3d(i, sec=sec)
                r3dsoma[1] += h.y3d(i, sec=sec)
                r3dsoma[2] += h.z3d(i, sec=sec)

        r3dsoma /= n3dsoma
        return r3dsoma

    def calc_seg_coords(self):
        """Calculate segment coordinates from 3d point coordinates"""
        ix = 0  # segment index

        p3dsoma = self.get_soma_pos()
        self.psoma = p3dsoma
        
        p0 = np.zeros((3, self.nseg))  # hold the coordinates of segment starting points
        p1 = np.zeros((3, self.nseg))  # hold the coordinates of segment end points
        p05 = np.zeros((3, self.nseg))
        d0 = np.zeros(self.nseg)
        d1 = np.zeros(self.nseg)

        for sec in self.hobj.all:
            n3d = int(h.n3d(sec=sec))  # get number of n3d points in each section
            p3d = np.zeros((3, n3d))  # to hold locations of 3D morphology for the current section
            l3d = np.zeros(n3d)  # to hold locations of 3D morphology for the current section
            diam3d = np.zeros(n3d)  # to diameters

            for i in range(n3d):
                p3d[0, i] = h.x3d(i, sec=sec) - p3dsoma[0]
                p3d[1, i] = h.y3d(i, sec=sec) - p3dsoma[1]  # shift coordinates such to place soma at the origin.
                p3d[2, i] = h.z3d(i, sec=sec) - p3dsoma[2]
                diam3d[i] = h.diam3d(i, sec=sec)
                l3d[i] = h.arc3d(i, sec=sec)

            l3d /= sec.L                  # normalize
            nseg = sec.nseg
            
            l0 = np.zeros(nseg)     # keep range of segment starting point 
            l1 = np.zeros(nseg)     # keep range of segment ending point
            l05 = np.zeros(nseg)
            
            for iseg, seg in enumerate(sec):
                l0[iseg] = seg.x - 0.5*1/nseg   # x (normalized distance along the section) for the beginning of the segment
                l1[iseg] = seg.x + 0.5*1/nseg   # x for the end of the segment
                l05[iseg] = seg.x

            p0[0, ix:ix+nseg] = np.interp(l0, l3d, p3d[0, :])
            p0[1, ix:ix+nseg] = np.interp(l0, l3d, p3d[1, :])
            p0[2, ix:ix+nseg] = np.interp(l0, l3d, p3d[2, :])
            d0[ix:ix+nseg] = np.interp(l0, l3d, diam3d[:])

            p1[0, ix:ix+nseg] = np.interp(l1, l3d, p3d[0, :])
            p1[1, ix:ix+nseg] = np.interp(l1, l3d, p3d[1, :])
            p1[2, ix:ix+nseg] = np.interp(l1, l3d, p3d[2, :])
            d1[ix:ix+nseg] = np.interp(l1, l3d, diam3d[:])

            p05[0,ix:ix+nseg] = np.interp(l05, l3d, p3d[0,:])
            p05[1,ix:ix+nseg] = np.interp(l05, l3d, p3d[1,:])
            p05[2,ix:ix+nseg] = np.interp(l05, l3d, p3d[2,:])

            ix += nseg
        self.seg_coords = {}

        self.seg_coords['p0'] = p0
        self.seg_coords['p1'] = p1
        self.seg_coords['p05'] = p05

        self.seg_coords['d0'] = d0
        self.seg_coords['d1'] = d1

        return self.seg_coords

    def set_seg_props(self):
        """Set segment properties which are invariant for all cell using this morphology"""
        seg_type = []
        seg_area = []
        seg_x = []
        seg_dist = []
        seg_length = []

        h.distance(sec=self.hobj.soma[0])   # measure distance relative to the soma
        
        for sec in self.hobj.all:
            fullsecname = sec.name()
            sec_type = fullsecname.split(".")[1][:4] # get sec name type without the cell name
            sec_type_swc = self.sec_type_swc[sec_type]  # convert to swc code

            for seg in sec:

                seg_area.append(h.area(seg.x))
                seg_x.append(seg.x)
                seg_length.append(sec.L/sec.nseg)
                seg_type.append(sec_type_swc)           # record section type in a list
                # seg_dist.append(h.distance(seg.x))  # distance to the center of the segment
                seg_dist.append(h.distance(seg))

        self.seg_prop = {}
        self.seg_prop['type'] = np.array(seg_type)
        self.seg_prop['area'] = np.array(seg_area)
        self.seg_prop['x'] = np.array(seg_x)
        self.seg_prop['dist'] = np.array(seg_dist)
        self.seg_prop['length'] = np.array(seg_length)
        self.seg_prop['dist0'] = self.seg_prop['dist'] - self.seg_prop['length']/2
        self.seg_prop['dist1'] = self.seg_prop['dist'] + self.seg_prop['length']/2

    def get_target_segments(self, edge_type):
        # Determine the target segments and their probabilities of connections for each new edge-type. Save the
        # information for each additional time a given edge-type is used on this morphology
        # TODO: Don't rely on edge-type-table, just use the edge?
        if edge_type in self._segments:
            return self._segments[edge_type]

        else:
            tar_seg_ix, tar_seg_prob = self.find_sections(edge_type.target_sections, edge_type.target_distance)
            self._segments[edge_type] = (tar_seg_ix, tar_seg_prob)
            return tar_seg_ix, tar_seg_prob

        """
        tar_sec_labels = edge_type.target_sections
        drange = edge_type.target_distance
        dmin, dmax = drange[0], drange[1]

        seg_d0 = self.seg_prop['dist0']  # use a more compact variables
        seg_d1 = self.seg_prop['dist1']
        seg_length = self.seg_prop['length']
        seg_area = self.seg_prop['area']
        seg_type = self.seg_prop['type']

        # Find the fractional overlap between the segment and the distance range:
        # this is done by finding the overlap between [d0,d1] and [dmin,dmax]
        # np.minimum(seg_d1,dmax) find the smaller of the two end locations
        # np.maximum(seg_d0,dmin) find the larger of the two start locations
        # np.maximum(0,overlap) is used to return zero when segments do not overlap
        # and then dividing by the segment length
        frac_overlap = np.maximum(0, (np.minimum(seg_d1, dmax) - np.maximum(seg_d0, dmin))) / seg_length
        ix_drange = np.where(frac_overlap > 0)  # find indexes with non-zero overlap
        ix_labels = np.array([], dtype=np.int)

        for tar_sec_label in tar_sec_labels:  # find indexes within sec_labels
            sec_type = self.sec_type_swc[tar_sec_label]  # get swc code for the section label
            ix_label = np.where(seg_type == sec_type)
            ix_labels = np.append(ix_labels, ix_label)  # target segment indexes

        tar_seg_ix = np.intersect1d(ix_drange, ix_labels)  # find intersection between indexes for range and labels
        tar_seg_length = seg_length[tar_seg_ix] * frac_overlap[tar_seg_ix]  # weighted length of targeted segments
        tar_seg_prob = tar_seg_length / np.sum(tar_seg_length)  # probability of targeting segments

        self._segments[edge_type] = (tar_seg_ix, tar_seg_prob)
        return tar_seg_ix, tar_seg_prob
        """

    def find_sections(self, target_sections, distance_range):
        dmin, dmax = distance_range[0], distance_range[1]

        seg_d0 = self.seg_prop['dist0']  # use a more compact variables
        seg_d1 = self.seg_prop['dist1']
        seg_length = self.seg_prop['length']
        seg_area = self.seg_prop['area']
        seg_type = self.seg_prop['type']

        # Find the fractional overlap between the segment and the distance range:
        # this is done by finding the overlap between [d0,d1] and [dmin,dmax]
        # np.minimum(seg_d1,dmax) find the smaller of the two end locations
        # np.maximum(seg_d0,dmin) find the larger of the two start locations
        # np.maximum(0,overlap) is used to return zero when segments do not overlap
        # and then dividing by the segment length
        frac_overlap = np.maximum(0, (np.minimum(seg_d1, dmax) - np.maximum(seg_d0, dmin))) / seg_length
        ix_drange = np.where(frac_overlap > 0)  # find indexes with non-zero overlap
        ix_labels = np.array([], dtype=np.int)

        for tar_sec_label in target_sections:  # find indexes within sec_labels
            sec_type = self.sec_type_swc[tar_sec_label]  # get swc code for the section label
            ix_label = np.where(seg_type == sec_type)
            ix_labels = np.append(ix_labels, ix_label)  # target segment indexes

        tar_seg_ix = np.intersect1d(ix_drange, ix_labels)  # find intersection between indexes for range and labels
        tar_seg_length = seg_length[tar_seg_ix] * frac_overlap[tar_seg_ix]  # weighted length of targeted segments
        tar_seg_prob = tar_seg_length / np.sum(tar_seg_length)  # probability of targeting segments
        return tar_seg_ix, tar_seg_prob


# Copyright 2017. Allen Institute. All rights reserved
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
# disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
# disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
import os
import numpy as np
from neuron import h

from bmtk.simulator.core.simulator_network import SimNetwork
from bmtk.simulator.core import sonata_reader
from bmtk.simulator.bionet.biocell import BioCell, BioCellSpontSyn
from bmtk.simulator.bionet.pointprocesscell import PointProcessCell, PointProcessCellSpontSyns
from bmtk.simulator.bionet.pointsomacell import PointSomaCell
from bmtk.simulator.bionet.virtualcell import VirtualCell
from bmtk.simulator.bionet.morphology import Morphology
from bmtk.simulator.bionet.io_tools import io
from bmtk.simulator.bionet import nrn
from bmtk.simulator.bionet.sonata_adaptors import BioNodeAdaptor, BioEdgeAdaptor
from .gids import GidPool

# TODO: leave this import, it will initialize some of the default functions for building neurons/synapses/weights.
import bmtk.simulator.bionet.default_setters


pc = h.ParallelContext()  # object to access MPI methods
MPI_size = int(pc.nhost())
MPI_rank = int(pc.id())


class BioNetwork(SimNetwork):
    model_type_col = 'model_type'

    def __init__(self):
        super(BioNetwork, self).__init__()
        self._io = io

        # TODO: Find a better way that will allow users to register their own class
        self._model_type_map = {
            'biophysical': BioCell,
            'point_process': PointProcessCell,
            'point_neuron': PointProcessCell,
            'point_soma': PointSomaCell,
            'virtual': VirtualCell
        }

        self._morphologies_cache = {}
        self._morphology_lookup = {}

        self._rank_node_gids = {}
        self._rank_node_ids = {}
        self._rank_nodes_by_model = {m_type: {} for m_type in self._model_type_map.keys()}
        self._remote_node_cache = {}
        self._virtual_nodes = {}
        self._disconnected_source_cells = {}

        self._cells_built = False
        self._connections_initialized = False

        self._gid_pool = GidPool()

        self.has_spont_syns = False
        self.spont_syns_filter = None
        self.spont_syns_times = None

    @property
    def gid_pool(self):
        return self._gid_pool

    @property
    def py_function_caches(self):
        return nrn

    def set_spont_syn_activity(self, precell_filter, timestamps):
        self._model_type_map = {
            'biophysical': BioCellSpontSyn,
            'point_process': PointProcessCellSpontSyns,
            'point_neuron': PointProcessCellSpontSyns,
            'point_soma': PointSomaCell,
            'virtual': VirtualCell
        }

        self.has_spont_syns = True
        self.spont_syns_filter = precell_filter
        self.spont_syns_times = timestamps

    def get_node_id(self, population, node_id):
        if node_id in self._rank_node_ids[population]:
            return self._rank_node_ids[population][node_id].node

        elif node_id in self._remote_node_cache[population]:
            return self._remote_node_cache[population][node_id]

        else:
            node_pop = self.get_node_population(population)
            node = node_pop.get_node(node_id)
            self._remote_node_cache[population][node_id] = node
            return node

    def cell_type_maps(self, model_type):
        return self._rank_nodes_by_model[model_type]

    def get_cell_node_id(self, population, node_id):
        return self._rank_node_ids[population].get(node_id, None)

    def get_cell_gid(self, gid):
        return self._rank_node_gids[gid]

    def get_local_cells(self):
        return self._rank_node_gids

    @property
    def local_gids(self):
        return list(self._rank_node_gids.keys())

    def add_nodes(self, node_population):
        self._gid_pool.add_pool(node_population.name, node_population.n_nodes())
        super(BioNetwork, self).add_nodes(node_population)

    def get_virtual_cells(self, population, node_id, spike_trains):
        if node_id in self._virtual_nodes[population]:
            return self._virtual_nodes[population][node_id]
        else:
            node = self.get_node_id(population, node_id)
            virt_cell = VirtualCell(node, population, spike_trains)
            self._virtual_nodes[population][node_id] = virt_cell
            return virt_cell

    def get_disconnected_cell(self, population, node_id, spike_trains):
        if population not in self._disconnected_source_cells:
            self._disconnected_source_cells[population] = {}

        if node_id in self._disconnected_source_cells[population]:
            virt_cell = self._disconnected_source_cells[population][node_id]
        else:
            node = self.get_node_id(population, node_id)
            virt_cell = VirtualCell(node, population, spike_trains)
            self._disconnected_source_cells[population][node_id] = virt_cell

        return virt_cell

    def _build_cell(self, bionode, population_name):
        if bionode.model_type in self._model_type_map:
            cell = self._model_type_map[bionode.model_type](bionode, population_name=population_name, bionetwork=self)
            self._rank_nodes_by_model[bionode.model_type][cell.gid] = cell
            return cell
        else:
            self.io.log_exception('Unrecognized model_type {}.'.format(bionode.model_type))

    def _register_adaptors(self):
        super(BioNetwork, self)._register_adaptors()
        self._node_adaptors['sonata'] = BioNodeAdaptor
        self._edge_adaptors['sonata'] = BioEdgeAdaptor

    def build_nodes(self):
        for node_pop in self.node_populations:
            self._remote_node_cache[node_pop.name] = {}
            node_ids_map = {}
            if node_pop.internal_nodes_only:
                for node in node_pop[MPI_rank::MPI_size]:
                    cell = self._build_cell(bionode=node, population_name=node_pop.name)
                    node_ids_map[node.node_id] = cell
                    self._rank_node_gids[cell.gid] = cell

            elif node_pop.mixed_nodes:
                # node population contains both internal and virtual (external) nodes and the virtual nodes must be
                # filtered out
                self._virtual_nodes[node_pop.name] = {}
                for node in node_pop[MPI_rank::MPI_size]:
                    if node.model_type == 'virtual':
                        continue
                    else:
                        cell = self._build_cell(bionode=node, population_name=node_pop.name)
                        node_ids_map[node.node_id] = cell


                        self._rank_node_gids[cell.gid] = cell

            elif node_pop.virtual_nodes_only:
                self._virtual_nodes[node_pop.name] = {}

            self._rank_node_ids[node_pop.name] = node_ids_map

        self.make_morphologies()
        self.set_seg_props()  # set segment properties by creating Morphologies
        self.calc_seg_coords()  # use for computing the ECP
        self._cells_built = True
        self.io.barrier()

    def set_seg_props(self):
        """Set morphological properties for biophysically (morphologically) detailed cells"""
        for _, morphology in self._morphologies_cache.items():
            morphology.set_seg_props()

    def calc_seg_coords(self):
        """Needed for the ECP calculations"""
        # TODO: Is there any reason this function can't be moved to make_morphologies()
        for morphology_file, morphology in self._morphologies_cache.items():
            morph_seg_coords = morphology.calc_seg_coords()   # needed for ECP calculations

            for gid in self._morphology_lookup[morphology_file]:
                self.get_cell_gid(gid).calc_seg_coords(morph_seg_coords)

    def make_morphologies(self):
        """Creating a Morphology object for each biophysical model"""
        # TODO: Let Morphology take care of the cache
        # TODO: Let other types have morphologies
        # TODO: Get all available morphologies from TypesTable or group
        for gid, cell in self._rank_node_gids.items():
            if not isinstance(cell, BioCell):
                continue

            morphology_file = cell.morphology_file
            if morphology_file in self._morphologies_cache:
                # create a single morphology object for each model_group which share that morphology
                morph = self._morphologies_cache[morphology_file]

                # associate morphology with a cell
                cell.set_morphology(morph)
                self._morphology_lookup[morphology_file].append(cell.gid)

            else:
                hobj = cell.hobj  # get hoc object (hobj) from the first cell with a new morphologys
                morph = Morphology(hobj)

                # associate morphology with a cell
                cell.set_morphology(morph)

                # create a single morphology object for each model_group which share that morphology
                self._morphologies_cache[morphology_file] = morph
                self._morphology_lookup[morphology_file] = [cell.gid]

        self.io.barrier()

    def _init_connections(self):
        if not self._connections_initialized:
            for gid, cell in self._rank_node_gids.items():
                cell.init_connections()
            self._connections_initialized = True

    def get_gj_id(self, network, src_nid, trg_nid, source_gap):
        '''
        Returns the gap junction id for the given nodes on a given network.

        :param source_gap: whether to return the id of the gap junction on the source node or the target node.
        '''
        if src_nid == trg_nid:
            raise Exception("Cells cannot have gap junctions with themselves.")

        gap_ids = self._gap_juncs[network]

        loc = np.where(np.logical_and(gap_ids["source_ids"] == src_nid, gap_ids["target_ids"] == trg_nid))[0]

        if len(loc) > 1:
            raise Exception("The gap junction file has more than one gap junction with the same ids.")
        elif len(loc) == 0:
            raise Exception("The gap junction file does not contain a gap junction with source id " +
             str(src_nid) + " and target id " + str(trg_nid) + " on network " + network)

        ids = [gap_ids["src_gap_ids"][loc[0]], gap_ids["trg_gap_ids"][loc[0]]]

        if not source_gap:
            ids.reverse()
        
        return ids

    def build_recurrent_edges(self):
        recurrent_edge_pops = [ep for ep in self._edge_populations if not ep.virtual_connections]
        if not recurrent_edge_pops:
            return

        self._init_connections()
        for edge_pop in recurrent_edge_pops:
            if edge_pop.recurrent_connections:
                source_population = edge_pop.source_nodes
                for trg_nid, trg_cell in self._rank_node_ids[edge_pop.target_nodes].items():
                    for edge in edge_pop.get_target(trg_nid):
                        src_node = self.get_node_id(source_population, edge.source_node_id)

                        if edge.is_gap_junction:
                            if source_population != edge_pop.target_nodes:
                                raise Exception("Gap junctions must be from the same network builder")
                            gj_ids = self.get_gj_id(source_population, edge.source_node_id, trg_nid, False)
                            trg_cell.set_syn_connection(edge, src_node, gj_ids=gj_ids)
                        else:
                            trg_cell.set_syn_connection(edge, src_node)
                for src_nid, src_cell in self._rank_node_ids[source_population].items():
                    for edge in edge_pop.get_source(src_nid):
                        if edge.is_gap_junction:
                            if source_population != edge_pop.target_nodes:
                                raise Exception("Gap junctions must be from the same network builder")

                            trg_node = self.get_node_id(edge_pop.target_nodes, edge.target_node_id)

                            gj_ids = self.get_gj_id(source_population, src_nid, edge.target_node_id, True)

                            src_cell.set_syn_connection(edge, trg_node, gj_ids=gj_ids)

            elif edge_pop.mixed_connections:
                # When dealing with edges that contain both virtual and recurrent edges we have to check every source
                # node to see if is virtual (bc virtual nodes can't be built yet). This conditional can significantly
                # slow down build time so we use a special loop that can be ignored.
                source_population = edge_pop.source_nodes
                for trg_nid, trg_cell in self._rank_node_ids[edge_pop.target_nodes].items():
                    for edge in edge_pop.get_target(trg_nid):
                        src_node = self.get_node_id(source_population, edge.source_node_id)
                        if src_node.model_type == 'virtual':
                            continue
                        trg_cell.set_syn_connection(edge, src_node)

        self.io.barrier()

    def build_replay_inputs(self, spike_trains, edges_path, edge_types_path, source_node_set, target_node_set):
        self._init_connections()

        src_cells = self.get_node_set(source_node_set)
        valid_src_ids = set(src_cells.gids())
        trg_cells = self.get_node_set(target_node_set)
        valid_trg_ids = set(trg_cells.gids())

        edges_pop_list = sonata_reader.load_edges(edges_path, edge_types_path, adaptor=self.get_edge_adaptor('sonata'))
        for edges_pop in edges_pop_list:
            edges_pop.initialize(self)

            trg_population = edges_pop.target_nodes
            src_population = edges_pop.source_nodes
            if src_population not in src_cells.population_names() or trg_population not in trg_cells.population_names():
                continue

            for trg_nid, trg_cell in self._rank_node_ids[trg_population].items():
                if trg_nid not in valid_trg_ids:
                    continue

                for edge in edges_pop.get_target(trg_nid):
                    src_nid = edge.source_node_id
                    if src_nid not in valid_src_ids:
                        continue

                    src_cell = self.get_disconnected_cell(src_population, src_nid, spike_trains)
                    trg_cell.set_syn_connection(edge, src_cell, src_cell)

        self.io.barrier()

    def find_edges(self, source_nodes=None, target_nodes=None):
        selected_edges = self._edge_populations[:]

        if source_nodes is not None:
            selected_edges = [edge_pop for edge_pop in selected_edges if edge_pop.source_nodes == source_nodes]

        if target_nodes is not None:
            selected_edges = [edge_pop for edge_pop in selected_edges if edge_pop.target_nodes == target_nodes]

        return selected_edges

    def add_spike_trains(self, spike_trains, node_set):
        self._init_connections()

        src_nodes = [node_pop for node_pop in self.node_populations if node_pop.name in node_set.population_names()]
        for src_node_pop in src_nodes:
            source_population = src_node_pop.name
            for edge_pop in self.find_edges(source_nodes=source_population):
                if edge_pop.virtual_connections:
                    for trg_nid, trg_cell in self._rank_node_ids[edge_pop.target_nodes].items():
                        for edge in edge_pop.get_target(trg_nid):
                            src_cell = self.get_virtual_cells(source_population, edge.source_node_id, spike_trains)
                            trg_cell.set_syn_connection(edge, src_cell, src_cell)

                elif edge_pop.mixed_connections:
                    raise NotImplementedError()

        self.io.barrier()


# Copyright 2017. Allen Institute. All rights reserved
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
# disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
# disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
from neuron import h


class IClamp(object):
    def __init__(self, amplitude, delay, duration, section_name='soma', section_index=0, section_dist=0.5):
        self._iclamp_amp = amplitude
        self._iclamp_del = delay
        self._iclamp_dur = duration

        # Determine what neuronal struction (axon, soma, dend, etc.), section/branch index, and where along the branch
        # it will occur. default is center of first soma section; hobj.soma[0](0.5)
        self._section_name = section_name
        self._section_index = section_index
        self._section_dist = section_dist
        self._stim = None

    def attach_current(self, cell):
        # self._stim = h.IClamp(cell.hobj.soma[0](0.5))
        hboj_sec = getattr(cell.hobj, self._section_name)[self._section_index](self._section_dist)
        self._stim = h.IClamp(hboj_sec)
        self._stim.delay = self._iclamp_del
        self._stim.dur = self._iclamp_dur
        self._stim.amp = self._iclamp_amp
        return self._stim


class FileIClamp(object):
    def __init__(self, amplitudes, dt):
        self._iclamp_amps = amplitudes
        self._iclamp_dt = dt
        self._stim = None

    def attach_current(self, cell):
        self._stim = h.IClamp(cell.hobj.soma[0](0.5))

        # Listed as necessary values in the docs to use play() with an IClamp.
        self._stim.delay = 0
        self._stim.dur = 1e9

        self._vect_stim = h.Vector(self._iclamp_amps)
        self._vect_stim.play(self._stim._ref_amp, self._iclamp_dt)  #Plays the amps to the IClamp amp variable with a given dt.

        return self._stim




import os
import numpy as np
from sympy.abc import x as symbolic_x
from sympy.abc import y as symbolic_y

from .linearfilter import SpatioTemporalFilter
from .spatialfilter import GaussianSpatialFilter
from .temporalfilter import TemporalFilterCosineBump
from .movie import Movie
from .lgnmodel1 import LGNModel, heat_plot
from .transferfunction import MultiTransferFunction, ScalarTransferFunction
from .lnunit import LNUnit, MultiLNUnit


class OnUnit(LNUnit):
    def __init__(self, linear_filter, transfer_function):
        assert linear_filter.amplitude > 0
        super(OnUnit, self).__init__(linear_filter, transfer_function)


class OffUnit(LNUnit):
    def __init__(self, linear_filter, transfer_function):
        assert linear_filter.amplitude < 0
        super(OffUnit, self).__init__(linear_filter, transfer_function)


class LGNOnOffCell(MultiLNUnit):
    """A cell model for a OnOff cell"""
    def __init__(self, on_filter, off_filter,
                 transfer_function=MultiTransferFunction((symbolic_x, symbolic_y),
                                                         'Heaviside(x)*(x)+Heaviside(y)*(y)')):
        """Summary

        :param on_filter:
        :param off_filter:
        :param transfer_function:
        """
        self.on_filter = on_filter
        self.off_filter = off_filter
        self.on_unit = OnUnit(self.on_filter, ScalarTransferFunction('s'))
        self.off_unit = OffUnit(self.off_filter, ScalarTransferFunction('s'))
        super(LGNOnOffCell, self).__init__([self.on_unit, self.off_unit], transfer_function)
        

class TwoSubfieldLinearCell(MultiLNUnit):
    def __init__(self, dominant_filter, nondominant_filter, subfield_separation=10, onoff_axis_angle=45,
                 dominant_subfield_location=(30, 40),
                 transfer_function=MultiTransferFunction((symbolic_x, symbolic_y),
                                                         'Heaviside(x)*(x)+Heaviside(y)*(y)')):
        self.subfield_separation = subfield_separation
        self.onoff_axis_angle = onoff_axis_angle
        self.dominant_subfield_location = dominant_subfield_location
        self.dominant_filter = dominant_filter
        self.nondominant_filter = nondominant_filter
        self.transfer_function = transfer_function

        self.dominant_unit = LNUnit(self.dominant_filter, ScalarTransferFunction('s'),
                                    amplitude=self.dominant_filter.amplitude)
        self.nondominant_unit = LNUnit(self.nondominant_filter, ScalarTransferFunction('s'),
                                       amplitude=self.dominant_filter.amplitude)

        super(TwoSubfieldLinearCell, self).__init__([self.dominant_unit, self.nondominant_unit],
                                                    self.transfer_function)
              
        self.dominant_filter.spatial_filter.translate = self.dominant_subfield_location
        hor_offset = np.cos(self.onoff_axis_angle*np.pi/180.)*self.subfield_separation + self.dominant_subfield_location[0]
        vert_offset = np.sin(self.onoff_axis_angle*np.pi/180.)*self.subfield_separation + self.dominant_subfield_location[1]
        rel_translation = (hor_offset, vert_offset)
        self.nondominant_filter.spatial_filter.translate = rel_translation


"""
class LGNOnCell(OnUnit):
    def __init__(self, **kwargs):
        self.position = kwargs.pop('position', None)
        self.weights = kwargs.pop('weights', None)
        self.kpeaks = kwargs.pop('kpeaks', None)
        self.delays = kwargs.pop('delays', None)
        self.amplitude = kwargs.pop('amplitude', None)
        self.sigma = kwargs.pop('sigma', None)
        self.transfer_function_str = kwargs.pop('transfer_function_str', 's')  # 'Heaviside(s)*s')
        self.metadata = kwargs.pop('metadata', {})

        temporal_filter = TemporalFilterCosineBump(self.weights, self.kpeaks, self.delays)
        spatial_filter = GaussianSpatialFilter(translate=self.position, sigma=self.sigma,
                                               origin=(0, 0))  # all distances measured from BOTTOM LEFT
        spatiotemporal_filter = SpatioTemporalFilter(spatial_filter, temporal_filter, amplitude=self.amplitude)
        transfer_function = ScalarTransferFunction(self.transfer_function_str)
        self.unit = OnUnit(spatiotemporal_filter, transfer_function)


class LGNOffCell(OffUnit):
    def __init__(self, **kwargs):
        lattice_unit_center = kwargs.pop('lattice_unit_center', None)
        weights = kwargs.pop('weights', None)
        kpeaks = kwargs.pop('kpeaks', None)
        amplitude = kwargs.pop('amplitude', None)
        sigma = kwargs.pop('sigma', None)
        width = kwargs.pop('width', 5)
        transfer_function_str = kwargs.pop('transfer_function_str', 'Heaviside(s)*s')

        dxi = np.random.uniform(-width*1./2, width*1./2)
        dyi = np.random.uniform(-width*1./2, width*1./2)
        temporal_filter = TemporalFilterCosineBump(weights, kpeaks)
        spatial_filter = GaussianSpatialFilter(translate=(dxi, dyi), sigma=sigma, origin=lattice_unit_center)  # all distances measured from BOTTOM LEFT
        spatiotemporal_filter = SpatioTemporalFilter(spatial_filter, temporal_filter, amplitude=amplitude)
        transfer_function = ScalarTransferFunction(transfer_function_str)
        super(LGNOnCell, self).__init__(spatiotemporal_filter, transfer_function)
"""

if __name__ == "__main__":
    movie_file = '/data/mat/iSee_temp_shared/movies/TouchOfEvil.npy'
    m_data = np.load(movie_file, 'r')
    m = Movie(m_data[1000:], frame_rate=30.)
    
    # Create second cell:
    transfer_function = ScalarTransferFunction('s')
    temporal_filter = TemporalFilterCosineBump((0.4, -0.3), (20, 60))
    cell_list = []
    for xi in np.linspace(0, m.data.shape[2], 5):
        for yi in np.linspace(0, m.data.shape[1], 5):
            spatial_filter_on = GaussianSpatialFilter(sigma=(2, 2), origin=(0, 0), translate=(xi, yi))
            on_linear_filter = SpatioTemporalFilter(spatial_filter_on, temporal_filter, amplitude=20)
            spatial_filter_off = GaussianSpatialFilter(sigma=(4, 4), origin=(0, 0), translate=(xi, yi))
            off_linear_filter = SpatioTemporalFilter(spatial_filter_off, temporal_filter, amplitude=-20) 
            on_off_cell = LGNOnOffCell(on_linear_filter, off_linear_filter)
            cell_list.append(on_off_cell)
    
    lgn = LGNModel(cell_list)  # Here include a list of all cells
    y = lgn.evaluate(m, downsample=100)  # Does the filtering + non-linearity on movie object m
    heat_plot(y, interpolation='none', colorbar=True)



import os
import logging
from subprocess import call
from distutils.dir_util import copy_tree


logger = logging.getLogger(__name__)


def compile_mechanisms(mechanisms_dir):
    logger.info('Attempting to compile NEURON mechanims under "{}"'.format(mechanisms_dir))
    cwd = os.getcwd()
    try:
        os.chdir(os.path.join(mechanisms_dir))
        call(['nrnivmodl', 'modfiles'])
        logger.info('  Success.')
    except Exception as e:
        logger.error('  Was unable to compile mechanism in {}'.format(mechanisms_dir))
    os.chdir(cwd)


def copy_modfiles(mechanisms_dir, cached_dir=None):
    if not os.path.exists(mechanisms_dir):
        logger.info('Creating mechanisms directory {}'.format(mechanisms_dir))

    if cached_dir is None:
        local_path = os.path.dirname(os.path.realpath(__file__))
        cached_dir = os.path.join(local_path, '..', 'scripts/bionet/mechanisms')

    logger.info('Copying mod files from {} to {}'.format(cached_dir, mechanisms_dir))
    copy_tree(cached_dir, mechanisms_dir)



# Copyright 2017. Allen Institute. All rights reserved
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
# disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
# disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
import numpy as np
from bmtk.simulator.bionet import utils, nrn
from bmtk.simulator.bionet.cell import Cell
from bmtk.simulator.bionet.io_tools import io
import six

from neuron import h

pc = h.ParallelContext()    # object to access MPI methods

class ConnectionStruct(object):
    def __init__(self, edge_prop, src_node, syn, connector, is_virtual=False, is_gap_junc=False):
        self._src_node = src_node
        self._edge_prop = edge_prop
        self._syn = syn
        self._connector = connector
        self._is_gj = is_gap_junc
        self._is_virtual = is_virtual

    @property
    def is_virtual(self):
        return self._is_virtual

    @property
    def is_gap_junc(self):
        return self._is_gj

    @property
    def source_node(self):
        return self._src_node

    @property
    def syn_weight(self):
        if self.is_gap_junc:
            return self._connector.g
        else:
            return self._connector.weight[0]

    @syn_weight.setter
    def syn_weight(self, val):
        if self.is_gap_junc:
            self._connector.g = val
        else:
            self._connector.weight[0] = val


class BioCell(Cell):
    """Implemntation of a morphologically and biophysically detailed type cell.

    """
    def __init__(self, node, population_name, bionetwork):
        super(BioCell, self).__init__(node=node, population_name=population_name, network=bionetwork)

        # Set up netcon object that can be used to detect and communicate cell spikes.
        self.set_spike_detector(bionetwork.spike_threshold)

        self._morph = None
        self._seg_coords = {}

        # Determine number of segments and store a list of all sections.
        self._nseg = 0
        self.set_nseg(bionetwork.dL)
        self._secs = []
        self._secs_by_id = []
        self.set_sec_array()

        self._save_conn = False  # bionetwork.save_connection
        self._synapses = []
        self._gap_juncs = []
        self._syn_src_net = []
        self._src_gids = []
        self._syn_seg_ix = []
        self._syn_sec_x = []
        self._edge_type_ids = []
        self._segments = None
        self._connections = []

        # potentially used by ecp module
        self.im_ptr = None
        self.imVec = None

        # used by xstim module
        self.ptr2e_extracellular = None

        self.__extracellular_mech = False

    def set_spike_detector(self, spike_threshold):
        nc = h.NetCon(self.hobj.soma[0](0.5)._ref_v, None, sec=self.hobj.soma[0])  # attach spike detector to cell
        nc.threshold = spike_threshold     
        pc.cell(self.gid, nc)  # associate gid with spike detector

    def set_nseg(self, dL):
        """Define number of segments in a cell"""
        self._nseg = 0
        for sec in self.hobj.all:
            sec.nseg = 1 + 2 * int(sec.L/(2*dL))
            self._nseg += sec.nseg # get the total number of segments in the cell

    def calc_seg_coords(self, morph_seg_coords):
        """Update the segment coordinates (after rotations) for individual cells"""
        phi_y = self._node.rotation_angle_yaxis
        phi_z = self._node.rotation_angle_zaxis
        phi_x = self._node.rotation_angle_xaxis

        # Rotate cell
        # TODO: Rotations should follow as described in sonata (https://github.com/AllenInstitute/sonata/blob/master/docs/SONATA_DEVELOPER_GUIDE.md).
        #  Need someone with graphics experience to check they are being done correctly (I'm not sure atm).
        RotX = utils.rotation_matrix([1, 0, 0], phi_x)
        RotY = utils.rotation_matrix([0, 1, 0], phi_y)  # rotate segments around yaxis normal to pia
        RotZ = utils.rotation_matrix([0, 0, 1], phi_z)  # rotate segments around zaxis to get a proper orientation
        RotXYZ = np.dot(RotX, RotY.dot(RotZ))

        # rotated coordinates around z axis first then shift relative to the soma
        self._seg_coords['p0'] = self._pos_soma + np.dot(RotXYZ, morph_seg_coords['p0'])
        self._seg_coords['p1'] = self._pos_soma + np.dot(RotXYZ, morph_seg_coords['p1'])
        self._seg_coords['p05'] = self._pos_soma + np.dot(RotXYZ, morph_seg_coords['p05'])

    def get_seg_coords(self):
        return self._seg_coords

    @property
    def morphology_file(self):
        # TODO: Get from self._node.morphology_file
        return self._node.morphology_file

    @property
    def morphology(self):
        return self._morph

    @morphology.setter
    def morphology(self, morphology_obj):
        self.set_morphology(morphology_obj)

    def set_morphology(self, morphology_obj):
        self._morph = morphology_obj

    def get_sections(self):
        return self._secs_by_id
        #return self._secs

    def get_sections_id(self):
        return self._secs_by_id

    def get_section(self, sec_id):
        return self._secs[sec_id]

    def store_segments(self):
        self._segments = []
        for sec in self._secs_by_id:
            for seg in sec:
                self._segments.append(seg)

    def get_segments(self):
        return self._segments

    def set_sec_array(self):
        """Arrange sections in an array to be access by index"""
        secs = []  # build ref to sections
        self._secs_by_id = []
        for sec in self.hobj.all:
            self._secs_by_id.append(sec)
            for _ in sec:
                secs.append(sec)  # section to which segments belongs

        self._secs = np.array(secs)

    def set_syn_connection(self, edge_prop, src_node, stim=None, gj_ids=None):
        syn_weight = edge_prop.syn_weight(src_node=src_node, trg_node=self._node)

        if edge_prop.is_gap_junction:
            if gj_ids == None:
                raise Exception("Gap junctions must have a gap junction id passed to set_syn_connection.")

            self._edge_props.append(edge_prop)
            self._src_gids.append(src_node.node_id)

            if edge_prop.preselected_targets:
                return self._set_gap_junc_preselected(edge_prop, src_node, syn_weight, gj_ids)
            else:
                return self._set_gap_junc(edge_prop, src_node, syn_weight, gj_ids)
        else:
            if edge_prop.preselected_targets:
                self._edge_props.append(edge_prop)
                self._src_gids.append(src_node.node_id)
                return self._set_connection_preselected(edge_prop, src_node, syn_weight, stim)
            else:
                self._edge_props += [edge_prop]*edge_prop.nsyns
                self._src_gids += [src_node.node_id]*edge_prop.nsyns
                return self._set_connections(edge_prop, src_node, syn_weight, stim)

    def _set_gap_junc_preselected(self, edge_prop, src_node, syn_weight, gj_ids):
        if edge_prop.nsyns < 1:
            return 1

        sec_x = edge_prop.afferent_section_pos
        sec_id = edge_prop.afferent_section_id
        section = self._secs_by_id[sec_id]

        #Sets up the section to be connected to the gap junction.
        pc.source_var(section(0.5)._ref_v, gj_ids[1], sec=section)

        #Creates gap junction.
        try:
            gap_junc = h.Gap(0.5, sec=section)
        except:
            raise Exception("You need the gap.mod file to create gap junctions.")

        #Attaches the source section to the gap junction.
        pc.target_var(gap_junc, gap_junc._ref_vgap, gj_ids[0])

        gap_junc.g = syn_weight

        self._connections.append(ConnectionStruct(edge_prop, src_node, gap_junc, gap_junc, False, True))

        self._gap_juncs.append(gap_junc)
        self._edge_type_ids.append(edge_prop.edge_type_id)

        if self._save_conn:
            self._save_connection(src_gid=src_node.gid, src_net=src_node.network, sec_x=sec_x, seg_ix=sec_id,
                                  edge_type_id=edge_prop.edge_type_id)

        return 1

    def _set_gap_junc(self, edge_prop, src_node, syn_weight, gj_ids):
        if edge_prop.nsyns < 1:
            return 1

        tar_seg_ix, tar_seg_prob = self._morph.get_target_segments(edge_prop)
        nsyns = 1

        # choose nsyn elements from seg_ix with probability proportional to segment area
        seg_ix = self.prng.choice(tar_seg_ix, nsyns, p=tar_seg_prob)[0]
        sec = self._secs[seg_ix]  # section where synapases connect
        x = self._morph.seg_prop['x'][seg_ix]  # distance along the section where synapse connects, i.e., seg_x

        if edge_prop.nsyns > 1:
            print("Warning: The number of synapses passed in was greater than 1, but only one gap junction will be made.")

        #Sets up the section to be connected to the gap junction.
        pc.source_var(sec(0.5)._ref_v, gj_ids[1], sec=sec)

        #Creates gap junction.
        try:
            gap_junc = h.Gap(0.5, sec=sec)
        except:
            raise Exception("You need the gap.mod file to create gap junctions.")

        #Attaches the source section to the gap junction.
        pc.target_var(gap_junc, gap_junc._ref_vgap, gj_ids[0])

        gap_junc.g = syn_weight

        self._connections.append(ConnectionStruct(edge_prop, src_node, gap_junc, gap_junc, False, True))

        self._gap_juncs.append(gap_junc)
        self._edge_type_ids.append(edge_prop.edge_type_id)

        if self._save_conn:
            self._save_connection(src_gid=src_node.gid, src_net=src_node.network, sec_x=x, seg_ix=sec_ix,
                                  edge_type_id=edge_prop.edge_type_id)

    def _set_connection_preselected(self, edge_prop, src_node, syn_weight, stim=None):
        # TODO: synapses should be loaded by edge_prop.load_synapse
        sec_x = edge_prop.afferent_section_pos
        sec_id = edge_prop.afferent_section_id

        section = self._secs_by_id[sec_id]
        # section = self._secs[sec_id]
        delay = edge_prop['delay']
        synapse_fnc = nrn.py_modules.synapse_model(edge_prop['model_template'])
        syn = synapse_fnc(edge_prop['dynamics_params'], sec_x, section)

        if stim is not None:
            nc = h.NetCon(stim.hobj, syn)  # stim.hobj - source, syn - target
        else:
            src_gid = self._network.gid_pool.get_gid(name=src_node.population_name, node_id=src_node.node_id)
            nc = pc.gid_connect(src_gid, syn)

        nc.weight[0] = syn_weight
        nc.delay = delay
        self._connections.append(ConnectionStruct(edge_prop, src_node, syn, nc, stim is not None))

        self._netcons.append(nc)
        self._synapses.append(syn)
        self._edge_type_ids.append(edge_prop.edge_type_id)
        if self._save_conn:
            self._save_connection(src_gid=src_node.node_id, src_net=src_node.network, sec_x=sec_x, seg_ix=sec_id,
                                  edge_type_id=edge_prop.edge_type_id)

        return 1

    def _set_connections(self, edge_prop, src_node, syn_weight, stim=None):
        tar_seg_ix, tar_seg_prob = self._morph.get_target_segments(edge_prop)
        nsyns = edge_prop.nsyns

        if len(tar_seg_ix) == 0:
            msg = 'Could not find target synaptic location for edge-type {}, Please check target_section and/or distance_range properties'.format(edge_prop.edge_type_id)
            io.log_warning(msg, all_ranks=True, display_once=True)
            return 0

        segs_ix = self.prng.choice(tar_seg_ix, nsyns, p=tar_seg_prob)
        secs = self._secs[segs_ix]  # sections where synapases connect
        xs = self._morph.seg_prop['x'][segs_ix]  # distance along the section where synapse connects, i.e., seg_x

        # TODO: this should be done just once
        synapses = [edge_prop.load_synapses(x, sec) for x, sec in zip(xs, secs)]

        delay = edge_prop['delay']
        self._synapses.extend(synapses)

        # TODO: Don't save this if not needed
        self._edge_type_ids.extend([edge_prop.edge_type_id]*len(synapses))

        for syn in synapses:
            # connect synapses
            if stim:
                nc = h.NetCon(stim.hobj, syn)
            else:
                src_gid = self._network.gid_pool.get_gid(name=src_node.population_name, node_id=src_node.node_id)
                nc = pc.gid_connect(src_gid, syn)

            nc.weight[0] = syn_weight
            nc.delay = delay
            self.netcons.append(nc)

            self._connections.append(ConnectionStruct(edge_prop, src_node, syn, nc, stim is not None))

        return nsyns

    def connections(self):
        return self._connections

    def _save_connection(self, src_gid, src_net, sec_x, seg_ix, edge_type_id):
        self._src_gids.append(src_gid)
        self._syn_src_net.append(src_net)
        self._syn_sec_x.append(sec_x)
        self._syn_seg_ix.append(seg_ix)
        self._edge_type_id.append(edge_type_id)

    def get_connection_info(self):
        # TODO: There should be a more effecient and robust way to return synapse information.
        return [[self.gid, self._syn_src_gid[i], self.network_name, self._syn_src_net[i], self._syn_seg_ix[i],
                 self._syn_sec_x[i], self.netcons[i].weight[0], self.netcons[i].delay, self._edge_type_id[i], 0]
                for i in range(len(self._synapses))]

    def init_connections(self):
        super(BioCell, self).init_connections()
        self._synapses = []
        self._syn_src_gid = []
        self._syn_seg_ix = []
        self._syn_sec_x = []

    def __set_extracell_mechanism(self):
        if not self.__extracellular_mech:
            for sec in self.hobj.all:
                sec.insert('extracellular')
            self.__extracellular_mech = True

    def setup_ecp(self):
        self.im_ptr = h.PtrVector(self._nseg)  # pointer vector
        # used for gathering an array of  i_membrane values from the pointer vector
        self.im_ptr.ptr_update_callback(self.set_im_ptr)
        self.imVec = h.Vector(self._nseg)

        self.__set_extracell_mechanism()
        #for sec in self.hobj.all:
        #    sec.insert('extracellular')

    def setup_xstim(self, set_nrn_mechanism=True):
        self.ptr2e_extracellular = h.PtrVector(self._nseg)
        self.ptr2e_extracellular.ptr_update_callback(self.set_ptr2e_extracellular)

        # Set the e_extracellular mechanism for all sections on this hoc object
        if set_nrn_mechanism:
            self.__set_extracell_mechanism()
            #for sec in self.hobj.all:
            #    sec.insert('extracellular')

    def set_im_ptr(self):
        """Set PtrVector to point to the _ref_i_membrane_ parameter"""
        jseg = 0
        for sec in self.hobj.all:  
            for seg in sec:
                self.im_ptr.pset(jseg, seg._ref_i_membrane_)  # notice the underscore at the end
                jseg += 1

    def get_im(self):
        """Gather membrane currents from PtrVector into imVec (does not need a loop!)"""
        self.im_ptr.gather(self.imVec)
        # Warning: as_numpy() seems to fail with in neuron 7.4 for python 3
        # return self.imVec.as_numpy()  # (nA)
        return np.array(self.imVec)

    def set_ptr2e_extracellular(self):
        jseg = 0
        for sec in self.hobj.all:
            for seg in sec:
                self.ptr2e_extracellular.pset(jseg, seg._ref_e_extracellular)
                jseg += 1

    def set_e_extracellular(self, vext):
        self.ptr2e_extracellular.scatter(vext)

    def print_synapses(self):
        rstr = ''
        for i in six.moves.range(len(self._syn_src_gid)):
            rstr += '{}> <-- {} ({}, {}, {}, {})\n'.format(i, self._syn_src_gid[i], self.netcons[i].weight[0],
                                                           self.netcons[i].delay, self._syn_seg_ix[i],
                                                           self._syn_sec_x[i])
        return rstr


class BioCellSpontSyn(BioCell):
    """Special class that allows certain synapses to spontaneously fire (without spiking) at a specific time.
    """
    def __init__(self, node, population_name, bionetwork):
        super(BioCellSpontSyn, self).__init__(node, population_name=population_name, bionetwork=bionetwork)

        # Get the timestamp at which synapses
        self._syn_timestamps = bionetwork.spont_syns_times
        self._syn_timestamps = [self._syn_timestamps] if np.isscalar(self._syn_timestamps) else self._syn_timestamps
        self._spike_trains = h.Vector(self._syn_timestamps)
        self._vecstim = h.VecStim()
        self._vecstim.play(self._spike_trains)

        self._precell_filter = bionetwork.spont_syns_filter
        assert(isinstance(self._precell_filter, dict))

    def _matches_filter(self, src_node):
        """Check to see if the presynaptic cell matches the criteria specified"""
        for k, v in self._precell_filter.items():
            if isinstance(v, (list, tuple)):
                if src_node[k] not in v:
                    return False
            else:
                if src_node[k] != v:
                    return False
        return True

    def _set_connections(self, edge_prop, src_node, syn_weight, stim=None):
        tar_seg_ix, tar_seg_prob = self._morph.get_target_segments(edge_prop)
        src_gid = src_node.node_id
        nsyns = edge_prop.nsyns

        # choose nsyn elements from seg_ix with probability proportional to segment area
        segs_ix = self.prng.choice(tar_seg_ix, nsyns, p=tar_seg_prob)
        secs = self._secs[segs_ix]  # sections where synapases connect
        xs = self._morph.seg_prop['x'][segs_ix]  # distance along the section where synapse connects, i.e., seg_x

        synapses = [edge_prop.load_synapses(x, sec) for x, sec in zip(xs, secs)]

        delay = edge_prop['delay']
        self._synapses.extend(synapses)

        for syn in synapses:
            # connect synapses
            if stim:
                nc = h.NetCon(stim.hobj, syn)
            elif self._matches_filter(src_node):
                nc = h.NetCon(self._vecstim, syn)
            else:
                nc = pc.gid_connect(src_gid, syn)
                syn_weight = 0.0

            nc.weight[0] = syn_weight
            nc.delay = delay
            self.netcons.append(nc)

            self._connections.append(ConnectionStruct(edge_prop, src_node, syn, nc, stim is not None))

        return nsyns

    def _set_connection_preselected(self, edge_prop, src_node, syn_weight, stim=None):
        # TODO: synapses should be loaded by edge_prop.load_synapse
        sec_x = edge_prop.afferent_section_pos
        sec_id = edge_prop.afferent_section_id
        section = self._secs_by_id[sec_id]
        # section = self._secs[sec_id]
        delay = edge_prop['delay']
        synapse_fnc = nrn.py_modules.synapse_model(edge_prop['model_template'])
        syn = synapse_fnc(edge_prop['dynamics_params'], sec_x, section)

        if stim is not None:
            nc = h.NetCon(stim.hobj, syn)

        elif self._matches_filter(src_node):
            nc = h.NetCon(self._vecstim, syn)

        else:
            nc = pc.gid_connect(src_node.node_id, syn)
            syn_weight = 0.0

        nc.weight[0] = syn_weight
        nc.delay = delay
        self._connections.append(ConnectionStruct(edge_prop, src_node, syn, nc, stim is not None))

        self._netcons.append(nc)
        self._synapses.append(syn)
        self._edge_type_ids.append(edge_prop.edge_type_id)
        if self._save_conn:
            self._save_connection(src_gid=src_node.node_id, src_net=src_node.network, sec_x=sec_x, seg_ix=sec_id,
                                  edge_type_id=edge_prop.edge_type_id)

        return 1



import nest
import re


def get_version():
    """Trys to get NEST version major, minor, and patch (optional) of the current running version of nest. Will return
    as a list of ints [major, minor, patch], although patch may be None.

    :return: [major, minor, patch] if able to parse version, None if fails. 'patch' may be None value.
    """

    # Try to get the version string
    version_str = None
    try:
        # NEST 2.* uses .version() to get version string
        version_str = nest.version()
    except AttributeError:
        pass

    if version_str is None:
        try:
            # For NEST 3.1 it uses __version__ attribute to store string
            version_str = nest.__version__
            if version_str.upper() == 'UNKNOWN':
                return [3, None, None]

        except AttributeError:
            pass

    if version_str is None:
        return None

    # parse the version string to get major, minor and patch numbers
    try:
        version_pattern = re.compile(r'.*nest-(\d+)\.(\d+)(?:\.(\d+))?.*')
        m = re.match(version_pattern, version_str)
        n_groups = len(m.groups())
        ver_major = int(m.group(1))
        ver_minor = int(m.group(2))
        ver_patch = int(m.group(3)) if n_groups >= 3 and m.group(3) is not None else None
        return [ver_major, ver_minor, ver_patch]

    except (AttributeError, IndexError, ValueError, TypeError) as err:
        return None


nest_version = get_version()
if nest_version is None:
    # For now default to assume NEST 3.*.* is being used
    nest_version = [3, None, None]


NEST_SYNAPSE_MODEL_PROP = 'model' if nest_version[0] == 2 else 'synapse_model'
NEST_SPIKE_DETECTOR = 'spike_detector' if nest_version[0] == 2 else 'spike_recorder'



# Copyright 2017. Allen Institute. All rights reserved
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
# disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
# disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
from neuron import h
import numpy as np

from bmtk.simulator.bionet.io_tools import io


class VirtualCell(object):
    """Representation of a Virtual/External node"""

    def __init__(self, node, population, spike_train_dataset):
        # VirtualCell is currently not a subclass of bionet.Cell class b/c the parent has a bunch of properties that
        # just don't apply to a virtual cell. May want to make bionet.Cell more generic in the future.
        self._node_id = node.node_id
        self._node = node
        self._population = population
        self._hobj = None
        self._spike_train_dataset = spike_train_dataset
        self._train_vec = []
        self.set_stim(node, self._spike_train_dataset)
        
    @property
    def node_id(self):
        return self._node_id

    @property
    def hobj(self):
        return self._hobj

    def set_stim(self, stim_prop, spike_train):
        """Gets the spike trains for each individual cell."""
        spikes = spike_train.get_times(node_id=self.node_id)

        if spikes is None:
            spikes = []

        if np.any(np.array(spikes) < 0.0):
            # NRN will fail if VecStim contains negative spike-time, throw an exception and log info for user
            io.log_exception('spike train {} contains negative number, unable to run virtual cell in NEURON'.format(
                spikes
            ))

        spikes = np.sort(spikes)  # sort the spikes for NEURON, will throw a segfault if not sorted

        self._train_vec = h.Vector(spikes)
        vecstim = h.VecStim()
        vecstim.play(self._train_vec)
        self._hobj = vecstim

    def __getitem__(self, item):
        return self._node[item]





class NodesReader(object):
    def __init__(self):
        self._has_internal_nodes = False
        self._has_virtual_nodes = False

    @property
    def name(self):
        raise NotImplementedError()

    @property
    def internal_nodes_only(self):
        return self._has_internal_nodes and not self._has_virtual_nodes

    @property
    def virtual_nodes_only(self):
        return self._has_virtual_nodes and not self._has_internal_nodes

    @property
    def mixed_nodes(self):
        return self._has_internal_nodes and self._has_virtual_nodes

    def initialize(self, network):
        raise NotImplementedError()

    @classmethod
    def load(cls, **properties):
        raise NotImplementedError()


class EdgesReader(object):
    unknown = 0
    recurrent = 0
    virtual = 1
    mixed = 2

    def __init__(self):
        self._connection_type = -1

    @property
    def recurrent_connections(self):
        return self._connection_type == self.recurrent

    @property
    def virtual_connections(self):
        return self._connection_type == self.virtual

    @property
    def mixed_connections(self):
        return self._connection_type == self.mixed

    @property
    def source_nodes(self):
        raise NotImplementedError()

    @property
    def target_nodes(self):
        raise NotImplementedError()

    def set_connection_type(self, src_pop, trg_pop):
        if src_pop.internal_nodes_only and trg_pop.internal_nodes_only:
            self._connection_type = self.recurrent

        elif src_pop.virtual_nodes_only and trg_pop.internal_nodes_only:
            self._connection_type = self.virtual

        else:
            self._connection_type = self.mixed

    def initialize(self, network):
        raise NotImplementedError()




def listify(files):
    # TODO: change this to include any iterable datastructures (sets, panda sequences, etc)
    if not isinstance(files, (list, tuple)):
        return [files]
    else:
        return files



import os
import csv
import h5py
import numpy as np
from neuron import h

from .sim_module import SimulatorMod
from bmtk.simulator.bionet.biocell import BioCell
# from bmtk.simulator.bionet.io_tools import io
# from bmtk.simulator.bionet.pointprocesscell import PointProcessCell
from bmtk.utils.reports import CompartmentReport

try:
    # Check to see if h5py is built to run in parallel
    if h5py.get_config().mpi:
        MembraneRecorder = CompartmentReport  # cell_vars.CellVarRecorderParallel
    else:
        MembraneRecorder = CompartmentReport  # cell_vars.CellVarRecorder
except Exception as e:
    MembraneRecorder = CompartmentReport  # cell_vars.CellVarRecorder

pc = h.ParallelContext()
MPI_RANK = int(pc.id())
N_HOSTS = int(pc.nhost())


class NetconReport(SimulatorMod):
    def __init__(self, tmp_dir, file_name, variable_name, cells, sections='all', syn_type='Exp2Syn', buffer_data=True,
                 transform={}):
        """Module used for saving NEURON cell properities at each given step of the simulation.

        :param tmp_dir:
        :param file_name: name of h5 file to save variable.
        :param variables: list of cell variables to record
        :param gids: list of gids to to record
        :param sections:
        :param buffer_data: Set to true then data will be saved to memory until written to disk during each block, reqs.
        more memory but faster. Set to false and data will be written to disk on each step (default: True)
        """
        self._all_variables = list(variable_name)
        self._variables = list(variable_name)

        self._tmp_dir = tmp_dir

        self._file_name = file_name if os.path.isabs(file_name) else os.path.join(tmp_dir, file_name)
        self._all_gids = cells
        self._local_gids = []
        self._sections = sections

        self._var_recorder = None
        #self._var_recorder = MembraneRecorder(self._file_name, self._tmp_dir, self._all_variables,
        #                                      buffer_data=buffer_data, mpi_rank=MPI_RANK, mpi_size=N_HOSTS)

        self._virt_lookup = {}
        self._gid_lookup = {}
        self._sec_lookup = {}

        self._gid_list = []  # list of all gids that will have their variables saved
        self._data_block = {}  # table of variable data indexed by [gid][variable]
        self._block_step = 0  # time step within a given block
        self._object_lookup = {}
        self._syn_type = syn_type
        self._gid_map = None

    def _get_gids(self, sim):
        selected_gids = set(sim.net.get_node_set(self._all_gids).gids())
        self._local_gids = list(set(sim.local_gids) & selected_gids)

    def _save_sim_data(self, sim):
        self._var_recorder.tstart = 0.0
        self._var_recorder.tstop = sim.tstop
        self._var_recorder.dt = sim.dt

    def _get_syn_location(self, nc, cell):
        if isinstance(cell, BioCell):
            sec_x = nc.postloc()
            sec = h.cas()
            sec_id = self._sec_lookup[cell.gid][sec]  # cell.get_section_id(sec)
            h.pop_section()
            return sec_id, sec_x
        else:
            return -1, -1

    def initialize(self, sim):
        self._gid_map = sim.net.gid_pool
        self._get_gids(sim)

        self._var_recorder = MembraneRecorder(self._file_name, mode='w', variable=self._variables[0],
                                              buffer_size=sim.nsteps_block, tstart=0.0, tstop=sim.tstop, dt=sim.dt,
                                              n_steps=sim.n_steps)
        #self._save_sim_data(sim)

        for node_pop in sim.net.node_populations:
            pop_name = node_pop.name
            for node in node_pop[0::1]:
                if node.model_type != 'virtual':
                    self._gid_lookup[node.gid] = (pop_name, node.node_id)

        for gid, cell in sim.net.get_local_cells().items():
            trg_pop, trg_id = self._gid_lookup[gid]
            if isinstance(cell, BioCell):
                self._sec_lookup[gid] = {sec_name: sec_id for sec_id, sec_name in enumerate(cell.get_sections_id())}

        for gid in self._local_gids:
            pop_id = self._gid_map.get_pool_id(gid)
            sec_list = []
            seg_list = []
            src_list = []
            syn_objects = []

            cell = sim.net.get_cell_gid(gid)
            for nc in cell.netcons:
                synapse = nc.syn()
                if self._syn_type is None or synapse.hname().startswith(self._syn_type):
                    sec_id, seg_x = self._get_syn_location(nc, cell)
                    src_gid = int(nc.srcgid())
                    sec_list.append(sec_id)
                    seg_list.append(seg_x)
                    src_list.append(src_gid)
                    syn_objects.append(nc.syn())
                elif self._syn_type == 'netcon':
                    syn_objects.append(nc)

            if syn_objects:
                # self._var_recorder.add_cell(gid, sec_list, seg_list, src_ids=src_list, trg_ids=[gid]*len(src_list))
                self._var_recorder.add_cell(node_id=pop_id.node_id, population=pop_id.population, element_ids=sec_list,
                                            element_pos=seg_list, src_ids=src_list, trg_ids=[gid]*len(src_list))

                self._object_lookup[gid] = syn_objects

        # self._var_recorder.initialize(sim.n_steps, sim.nsteps_block)
        self._var_recorder.initialize()

    def step(self, sim, tstep):
        # save all necessary cells/variables at the current time-step into memory
        for gid, netcon_objs in self._object_lookup.items():
            pop_id = self._gid_map.get_pool_id(gid)
            for var_name in self._variables:
                syn_values = [getattr(syn, var_name) for syn in netcon_objs]
                if syn_values:
                    self._var_recorder.record_cell(pop_id.node_id, population=pop_id.population, vals=syn_values,
                                                   tstep=tstep)

        self._block_step += 1

    def block(self, sim, block_interval):
        # write variables in memory to file
        self._var_recorder.flush()

    def finalize(self, sim):
        # TODO: Build in mpi signaling into var_recorder
        pc.barrier()
        self._var_recorder.close()

        #pc.barrier()
        #self._var_recorder.merge()



# Copyright 2017. Allen Institute. All rights reserved
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
# disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
# disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#


class SimEdge(object):
    def __init__(self, original_params, dynamics_params):
        self._orig_params = original_params
        self._dynamics_params = dynamics_params
        self._updated_params = {'dynamics_params': self._dynamics_params}

    @property
    def edge_type_id(self):
        return self._orig_params['edge_type_id']

    def __getitem__(self, item):
        if item in self._updated_params:
            return self._updated_params[item]
        else:
            return self._orig_params[item]


class PopEdge(SimEdge):
    def __init__(self, source_pop, target_pop, edge_params, dynamics_params):
        super(PopEdge, self).__init__(edge_params, dynamics_params)
        self.__source_pop = source_pop
        self.__target_pop = target_pop
        self._weight = self.__get_prop('weight', 0.0)
        self._nsyns = self.__get_prop('nsyns', 0)
        self._delay = self.__get_prop('delay', 0.0)

    @property
    def source(self):
        return self.__source_pop

    @property
    def target(self):
        return self.__target_pop

    @property
    def params(self):
        return self._orig_params

    @property
    def weight(self):
        return self._weight

    @weight.setter
    def weight(self, value):
        self._weight = value

    @property
    def nsyns(self):
        return self._nsyns

    @nsyns.setter
    def nsyns(self, value):
        self._nsyns = value

    @property
    def delay(self):
        return self._delay

    @delay.setter
    def delay(self, value):
        self._delay = value

    def __get_prop(self, name, default=None):
        if name in self._orig_params:
            return self._orig_params[name]
        elif name in self._dynamics_params:
            return self._dynamics_params[name]
        else:
            return default

    def __repr__(self):
        relevant_params = "weight: {}, delay: {}, nsyns: {}".format(self.weight, self.delay, self.nsyns)
        rstr = "{} --> {} {{{}}}".format(self.source.pop_id, self.target.pop_id, relevant_params)
        return rstr



# Copyright 2017. Allen Institute. All rights reserved
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
# disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
# disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
import types
import numpy as np

from bmtk.simulator.core.sonata_reader import NodeAdaptor, SonataBaseNode, EdgeAdaptor, SonataBaseEdge
from bmtk.simulator.bionet import nrn


class BioNode(SonataBaseNode):
    @property
    def position(self):
        return self._prop_adaptor.position(self._node)

    @property
    def morphology_file(self):
        return self._node['morphology']

    @property
    def rotation_angle_xaxis(self):
        return self._prop_adaptor.rotation_angle_xaxis(self._node)

    @property
    def rotation_angle_yaxis(self):
        # TODO: Combine rotation alnges into a single property
        return self._prop_adaptor.rotation_angle_yaxis(self._node)

    @property
    def rotation_angle_zaxis(self):
        return self._prop_adaptor.rotation_angle_zaxis(self._node)

    @property
    def rotations(self):
        return self._prop_adaptor.rotations(self._node)

    @property
    def rotations_quaternion(self):
        return self._prop_adaptor.rotations(self._node)

    def load_cell(self):
        model_template = self.model_template
        template_name = model_template[1]
        model_type = self.model_type
        if nrn.py_modules.has_cell_model(self['model_template'], model_type):
            cell_fnc = nrn.py_modules.cell_model(self['model_template'], model_type)
        else:
            cell_fnc = nrn.py_modules.cell_model(model_template[0], model_type)

        dynamics_params = self.dynamics_params
        hobj = cell_fnc(self, template_name, dynamics_params)

        for model_processing_str in self.model_processing:
            processing_fnc = nrn.py_modules.cell_processor(model_processing_str)
            hobj = processing_fnc(hobj, self, dynamics_params)

        return hobj


class BioNodeAdaptor(NodeAdaptor):
    def get_node(self, sonata_node):
        return BioNode(sonata_node, self)

    @classmethod
    def patch_adaptor(cls, adaptor, node_group, network):
        node_adaptor = NodeAdaptor.patch_adaptor(adaptor, node_group, network)

        # Position
        if 'positions' in node_group.all_columns:
            node_adaptor.position = types.MethodType(positions, adaptor)
        elif 'position' in node_group.all_columns:
            node_adaptor.position = types.MethodType(position, adaptor)
        elif 'x' in node_group.all_columns:
            if 'z' in node_group.all_columns and 'y' in node_group.all_columns:
                node_adaptor.position = types.MethodType(position_xyz, adaptor)
            elif 'y' in node_group.all_columns:
                node_adaptor.position = types.MethodType(position_xy, adaptor)
            else:
                node_adaptor.position = types.MethodType(position_x, adaptor)
        else:
            node_adaptor.position = types.MethodType(positions_default, adaptor)

        # Rotation angles
        if 'rotation_angle_xaxis' in node_group.all_columns:
            node_adaptor.rotation_angle_xaxis = types.MethodType(rotation_angle_x, node_adaptor)
        else:
            node_adaptor.rotation_angle_xaxis = types.MethodType(rotation_angle_default, node_adaptor)

        if 'rotation_angle_yaxis' in node_group.all_columns:
            node_adaptor.rotation_angle_yaxis = types.MethodType(rotation_angle_y, node_adaptor)
        else:
            node_adaptor.rotation_angle_yaxis = types.MethodType(rotation_angle_default, node_adaptor)

        if 'rotation_angle_zaxis' in node_group.all_columns:
            node_adaptor.rotation_angle_zaxis = types.MethodType(rotation_angle_z, node_adaptor)
        else:
            node_adaptor.rotation_angle_zaxis = types.MethodType(rotation_angle_default, node_adaptor)

        if 'rotations' in node_group.all_columns:
            node_adaptor.rotations = types.MethodType(rotations, node_adaptor)
        else:
            node_adaptor.rotations = types.MethodType(value_none, node_adaptor)

        return node_adaptor


def rotations(self, node):
    return node['rotations']

def value_none(self, node):
    return None

def positions_default(self, node):
    return np.array([0.0, 0.0, 0.0])


def positions(self, node):
    return node['positions']


def position(self, node):
    return node['position']


def rotation_angle_default(self, node):
    return 0.0


def rotation_angle_x(self, node):
    return node['rotation_angle_xaxis']


def rotation_angle_y(self, node):
    return node['rotation_angle_yaxis']


def rotation_angle_z(self, node):
    return node['rotation_angle_zaxis']


def position_xyz(self, node):
    return np.array([node['x'], node['y'], node['z']])

def position_xy(self, node):
    return np.array([node['x'], node['y'], 0.0])

def position_x(self, node):
    return np.array([node['x'], 0.0, 0.0])

class BioEdge(SonataBaseEdge):
    def load_synapses(self, section_x, section_id):
        synapse_fnc = nrn.py_modules.synapse_model(self.model_template)
        return synapse_fnc(self.dynamics_params, section_x, section_id)

    @property
    def afferent_section_id(self):
        return self._prop_adaptor.aff_sec_id(self._edge)

    @property
    def afferent_section_pos(self):
        return self._prop_adaptor.aff_sec_pos(self._edge)



class BioEdgeAdaptor(EdgeAdaptor):
    def get_edge(self, sonata_edge):
        return BioEdge(sonata_edge, self)

    @classmethod
    def patch_adaptor(cls, adaptor, edge_group):
        edge_adaptor = EdgeAdaptor.patch_adaptor(adaptor, edge_group)

        if 'sec_id' in edge_group.all_columns:
            edge_adaptor.aff_sec_id = types.MethodType(aff_sec_id_old, adaptor)
        elif 'afferent_section_id' in edge_group.all_columns:
            edge_adaptor.aff_sec_id = types.MethodType(aff_sec_id, adaptor)
        else:
            edge_adaptor.aff_sec_id = types.MethodType(ret_none_function, adaptor)

        if 'sec_x' in edge_group.all_columns:
            edge_adaptor.aff_sec_pos = types.MethodType(aff_sec_pos_old, adaptor)
        elif 'afferent_section_pos' in edge_group.all_columns:
            edge_adaptor.aff_sec_pos = types.MethodType(aff_sec_pos, adaptor)
        else:
            edge_adaptor.aff_sec_pos = types.MethodType(ret_none_function, adaptor)

        return edge_adaptor


def ret_none_function(self, edge):
    return None


def aff_sec_id(self, edge):
    return edge['afferent_section_id']


def aff_sec_id_old(self, edge):
    return edge['sec_id']


def aff_sec_pos(self, edge):
    return edge['afferent_section_pos']


def aff_sec_pos_old(self, edge):
    return edge['sec_x']



import numpy as np
import scipy.interpolate as sinterp
import scipy.integrate as spi
import warnings
import scipy.optimize as sopt
import scipy.stats as sps


def generate_renewal_process(t0, t1, renewal_distribution):
    last_event_time = t0
    curr_interevent_time = float(renewal_distribution())
    event_time_list = []
    while last_event_time+curr_interevent_time <= t1:
        event_time_list.append(last_event_time+curr_interevent_time)
        curr_interevent_time = float(renewal_distribution())
        last_event_time = event_time_list[-1]
        
    return event_time_list 


def generate_poisson_process(t0, t1, rate):
    if rate is None:
        raise ValueError('Rate cannot be None')
    if rate > 10000:
        warnings.warn('Very high rate encountered: %s' % rate)

    try: 
        assert rate >= 0
    except AssertionError: 
        raise ValueError('Negative rate (%s) not allowed' % rate)
    
    try: 
        assert rate < np.inf
    except AssertionError: 
        raise ValueError('Rate (%s) must be finite' % rate) 

    if rate == 0:
        return []
    else:
        return generate_renewal_process(t0, t1, sps.expon(0, 1./rate).rvs)


def generate_inhomogenous_poisson(t_range, y_range, seed=None):
    if not seed == None: np.random.seed(seed) 
    spike_list = []
    for tl, tr, y in zip(t_range[:-1], t_range[1:], y_range[:-1]):
        spike_list += generate_poisson_process(tl, tr, y) 
    return spike_list


def generate_poisson_rescaling(t, y, seed=None):
    y = np.array(y)
    t = np.array(t)
    assert not np.any(y < 0)
    f = sinterp.interp1d(t, y, fill_value=0, bounds_error=False)
    return generate_poisson_rescaling_function(lambda y, t: f(t), t[0], t[-1], seed=seed)
    

def generate_poisson_rescaling_function(f, t_min, t_max, seed=None):
    def integrator(t0, t1):
        return spi.odeint(f, 0, [t0, t1])[1][0]
    
    if not seed == None:
        np.random.seed(seed)
    
    spike_train = []
    while t_min < t_max:
        e0 = np.random.exponential()

        def root_function(t):
            return e0 - integrator(t_min, t)
        
        try:
            with warnings.catch_warnings(record=True) as w:
                result = sopt.root(root_function, .1)
            assert result.success
        except AssertionError:
            if not e0 < integrator(t_min, t_max):
                assert Exception
            else:
                break

        t_min = result.x[0]
        spike_train.append(t_min)

    return np.array(spike_train)



# Copyright 2017. Allen Institute. All rights reserved
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
# disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
# disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
# products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
import os
import json
import ast
import numpy as np

# from bmtk.simulator.core.config import ConfigDict
from bmtk.simulator.core.simulation_config import SimulationConfig as ConfigDict
### from bmtk.simulator.utils.property_maps import NodePropertyMap, EdgePropertyMap
from bmtk.utils import sonata
from bmtk.simulator.core.io_tools import io

from bmtk.simulator.core.node_sets import NodeSet, NodeSetAll


"""
Creates a graph of nodes and edges from multiple network files for all simulators.

Consists of edges and nodes. All classes are abstract and should be reimplemented by a specific simulator. Also
contains base factor methods for building a network from a config file (or other).
"""

class SimGraph(object):
    model_type_col = 'model_type'

    def __init__(self):
        self._components = {}  # components table, i.e. paths to model files.
        self._io = io

        self._node_populations = {}
        self._internal_populations_map = {}
        self._virtual_populations_map = {}

        self._virtual_cells_nid = {}

        self._recurrent_edges = {}
        self._external_edges = {}

        self._node_sets = {}
        self._using_gids = False

    @property
    def io(self):
        return self._io

    @property
    def node_populations(self):
        return list(self._node_populations.keys())

    def get_node_set(self, node_set):
        if node_set in self._node_sets.keys():
            return self._node_sets[node_set]

        elif isinstance(node_set, (dict, list)):
            return NodeSet(node_set, self)

        else:
            self.io.log_exception('Unable to load or find node_set "{}"'.format(node_set))

    def get_node_populations(self):
        return self._node_populations.values()

    def get_node_population(self, population_name):
        return self._node_populations[population_name]

    def get_component(self, key):
        """Get the value of item in the components dictionary.

        :param key: name of component
        :return: value assigned to component
        """
        return self._components[key]

    def add_component(self, key, value):
        """Add a component key-value pair

        :param key: name of component
        :param value: value
        """
        self._components[key] = value

    def _validate_components(self):
        """Make sure various components (i.e. paths) exists before attempting to build the graph."""
        return True

    def __avail_model_types(self, population):
        model_types = set()
        for grp in population.groups:
            if self.model_type_col not in grp.all_columns:
                self.io.log_exception('model_type is missing from nodes.')

            model_types.update(set(np.unique(grp.get_values(self.model_type_col))))
        return model_types

    def _preprocess_node_types(self, node_population):
        # TODO: The following figures out the actually used node-type-ids. For mem and speed may be better to just
        # process them all
        node_type_ids = node_population.type_ids
        # TODO: Verify all the node_type_ids are in the table
        node_types_table = node_population.types_table

        # TODO: Convert model_type to a enum
        morph_dir = self.get_component('morphologies_dir')
        if morph_dir is not None and 'morphology' in node_types_table.columns:
            for nt_id in node_type_ids:
                node_type = node_types_table[nt_id]
                if node_type['morphology'] is None:
                    continue
                # TODO: Check the file exits
                # TODO: See if absolute path is stored in csv
                node_type['morphology'] = os.path.join(morph_dir, node_type['morphology'])

        if 'dynamics_params' in node_types_table.columns and 'model_type' in node_types_table.columns:
            for nt_id in node_type_ids:
                node_type = node_types_table[nt_id]
                dynamics_params = node_type['dynamics_params']
                if isinstance(dynamics_params, dict):
                    continue

                model_type = node_type['model_type']
                if model_type == 'biophysical':
                    params_dir = self.get_component('biophysical_neuron_models_dir')
                elif model_type == 'point_process':
                    params_dir = self.get_component('point_neuron_models_dir')
                elif model_type == 'point_soma':
                    params_dir = self.get_component('point_neuron_models_dir')
                else:
                    # Not sure what to do in this case, throw Exception?
                    params_dir = self.get_component('custom_neuron_models')

                params_path = os.path.join(params_dir, dynamics_params)

                # see if we can load the dynamics_params as a dictionary. Otherwise just save the file path and let the
                # cell_model loader function handle the extension.
                try:
                    params_val = json.load(open(params_path, 'r'))
                    node_type['dynamics_params'] = params_val
                except Exception:
                    # TODO: Check dynamics_params before
                    self.io.log_exception('Could not find node dynamics_params file {}.'.format(params_path))

    def _preprocess_edge_types(self, edge_pop):
        edge_types_table = edge_pop.types_table
        edge_type_ids = np.unique(edge_pop.type_ids)

        for et_id in edge_type_ids:
            edge_type = edge_types_table[et_id]
            if 'dynamics_params' in edge_types_table.columns:
                dynamics_params = edge_type['dynamics_params']
                params_dir = self.get_component('synaptic_models_dir')

                params_path = os.path.join(params_dir, dynamics_params)

                # see if we can load the dynamics_params as a dictionary. Otherwise just save the file path and let the
                # cell_model loader function handle the extension.
                try:
                    params_val = json.load(open(params_path, 'r'))
                    edge_type['dynamics_params'] = params_val
                except Exception:
                    # TODO: Check dynamics_params before
                    self.io.log_exception('Could not find edge dynamics_params file {}.'.format(params_path))

            # Split target_sections
            if 'target_sections' in edge_type:
                trg_sec = edge_type['target_sections']
                if trg_sec is not None:
                    try:
                        edge_type['target_sections'] = ast.literal_eval(trg_sec)
                    except Exception as exc:
                        self.io.log_warning('Unable to split target_sections list {}'.format(trg_sec))
                        edge_type['target_sections'] = None

            # Split target distances
            if 'distance_range' in edge_type:
                dist_range = edge_type['distance_range']
                if dist_range is not None:
                    try:
                        # TODO: Make the distance range has at most two values
                        edge_type['distance_range'] = json.loads(dist_range)
                    except Exception as e:
                        try:
                            edge_type['distance_range'] = [0.0, float(dist_range)]
                        except Exception as e:
                            self.io.log_warning('Unable to parse distance_range {}'.format(dist_range))
                            edge_type['distance_range'] = None

    def external_edge_populations(self, src_pop, trg_pop):
        return self._external_edges.get((src_pop, trg_pop), [])

    def add_nodes(self, sonata_file, populations=None):
        """Add nodes from a network to the graph.

        :param sonata_file: A NodesFormat type object containing list of nodes.
        :param populations: name/identifier of network. If none will attempt to retrieve from nodes object
        """
        nodes = sonata_file.nodes

        selected_populations = nodes.population_names if populations is None else populations
        for pop_name in selected_populations:
            if pop_name not in nodes:
                # when user wants to simulation only a few populations in the file
                print('HERE')
                continue

            if pop_name in self.node_populations:
                # Make sure their aren't any collisions
                self.io.log_exception('There are multiple node populations with name {}.'.format(pop_name))

            node_pop = nodes[pop_name]
            self._preprocess_node_types(node_pop)
            self._node_populations[pop_name] = node_pop

            # Segregate into virtual populations and non-virtual populations
            model_types = self.__avail_model_types(node_pop)
            if 'virtual' in model_types:
                self._virtual_populations_map[pop_name] = node_pop
                self._virtual_cells_nid[pop_name] = {}
                model_types -= set(['virtual'])
                if model_types:
                    # We'll allow a population to have virtual and non-virtual nodes but it is not ideal
                    self.io.log_warning('Node population {} contains both virtual and non-virtual nodes which can '.format(pop_name) +
                                        'cause memory and build-time inefficency. Consider separating virtual nodes ' +
                                        'into their own population')

            if model_types:
                self._internal_populations_map[pop_name] = node_pop

            self._node_sets[pop_name] = NodeSet({'population': pop_name}, self)

    def build_nodes(self):
        raise NotImplementedError

    def build_recurrent_edges(self):
        raise NotImplementedError

    def add_edges(self, sonata_file, populations=None, source_pop=None, target_pop=None):
        """

        :param sonata_file:
        :param populations:
        :param source_pop:
        :param target_pop:
        :return:
        """
        edges = sonata_file.edges
        selected_populations = edges.population_names if populations is None else populations

        for pop_name in selected_populations:
            if pop_name not in edges:
                continue

            edge_pop = edges[pop_name]
            self._preprocess_edge_types(edge_pop)

            # Check the source nodes exists
            src_pop = source_pop if source_pop is not None else edge_pop.source_population
            is_internal_src = src_pop in self._internal_populations_map.keys()
            is_external_src = src_pop in self._virtual_populations_map.keys()

            trg_pop = target_pop if target_pop is not None else edge_pop.target_population
            is_internal_trg = trg_pop in self._internal_populations_map.keys()

            if not is_internal_trg:
                self.io.log_exception(('Node population {} does not exists (or consists of only virtual nodes). ' +
                                      '{} edges cannot create connections.').format(trg_pop, pop_name))

            if not (is_internal_src or is_external_src):
                self.io.log_exception('Source node population {} not found. Please update {} edges'.format(src_pop,
                                                                                                           pop_name))
            if is_internal_src:
                if trg_pop not in self._recurrent_edges:
                    self._recurrent_edges[trg_pop] = []
                self._recurrent_edges[trg_pop].append(edge_pop)

            if is_external_src:
                if trg_pop not in self._external_edges:
                    self._external_edges[(src_pop, trg_pop)] = []
                self._external_edges[(src_pop, trg_pop)].append(edge_pop)

    @classmethod
    def from_config(cls, conf, **properties):
        """Generates a graph structure from a json config file or dictionary.

        :param conf: name of json config file, or a dictionary with config parameters
        :param properties: optional properties.
        :return: A graph object of type cls
        """
        graph = cls(**properties)

        # The simulation run script should create a config-dict since it's likely to vary based on the simulator engine,
        # however in the case the user doesn't we will try a generic conversion from dict/json to ConfigDict
        if isinstance(conf, ConfigDict):
            config = conf
        else:
            try:
                config = ConfigDict.load(conf)
            except Exception as e:
                graph.io.log_exception('Could not convert {} (type "{}") to json.'.format(conf, type(conf)))

        if not config.with_networks:
            graph.io.log_exception('Could not find any network files. Unable to build network.')

        # TODO: These are simulator specific
        graph.spike_threshold = config.spike_threshold
        graph.dL = config.dL

        # load components
        for name, value in config.components.items():
            graph.add_component(name, value)
        graph._validate_components()

        # load nodes
        gid_map = config.gid_mappings
        for node_dict in config.nodes:
            nodes_net = sonata.File(data_files=node_dict['nodes_file'], data_type_files=node_dict['node_types_file'],
                                    gid_table=gid_map)
            graph.add_nodes(nodes_net)

        # load edges
        for edge_dict in config.edges:
            target_network = edge_dict['target'] if 'target' in edge_dict else None
            source_network = edge_dict['source'] if 'source' in edge_dict else None
            edge_net = sonata.File(data_files=edge_dict['edges_file'], data_type_files=edge_dict['edge_types_file'])
            graph.add_edges(edge_net, source_pop=target_network, target_pop=source_network)

        graph._node_sets['all'] = NodeSetAll(graph)
        for ns_name, ns_filter in conf.node_sets.items():
            graph._node_sets[ns_name] = NodeSet(ns_filter, graph)

        return graph
